import base64
import logging
import pathlib
import stat
import typing
import urllib.parse

import pydantic
import sprockets.mixins.http
import tornado.web
import yarl

from imbi import errors, oauth2, version


class Namespace(pydantic.BaseModel):
    id: int


class ProjectLinks(pydantic.BaseModel):
    self: str


class ProjectInfo(pydantic.BaseModel):
    id: int
    default_branch: str
    namespace: Namespace
    web_url: str
    links: ProjectLinks = pydantic.Field(..., alias='_links')


PROJECT_DEFAULTS = {
    'default_branch': 'main',
    'issues_access_level': 'disabled',
    'issues_enabled': False,  # deprecated but required to disable :(
    'builds_access_level': 'enabled',
    'merge_requests_access_level': 'enabled',
    'operations_access_level': 'disabled',
    'packages_enabled': False,
    'pages_access_level': 'disabled',
    'snippets_access_level': 'enabled',
    'wiki_access_level': 'disabled',
    'wiki_enabled': False,  # deprecated but required to disable :(
}


class GitLabClient(sprockets.mixins.http.HTTPClientMixin):
    """API Client for GitLab.

    This client sends authenticated HTTP requests to the GitLab API.

    """
    def __init__(self,
                 token: oauth2.IntegrationToken,
                 application: tornado.web.Application):
        super().__init__()
        settings = application.settings['automations']['gitlab']
        self.logger = logging.getLogger(__package__).getChild('GitLabClient')
        self.restrict_to_user = settings.get('restrict_to_user', False)
        self.token = token
        self.headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.token.access_token}',
        }
        self._user_info: typing.Union[dict, None] = None
        self._user_namespace: typing.Union[dict, None] = None

    async def api(self, url: typing.Union[yarl.URL, str], *,
                  method: str = 'GET', **kwargs):
        """Make an authenticated request to the GitLab API."""
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            url = self.token.integration.api_endpoint / url.path.lstrip('/')
        request_headers = kwargs.setdefault('request_headers', {})
        request_headers.update(self.headers)
        if kwargs.get('body', None) is not None:
            request_headers['Content-Type'] = 'application/json'
            kwargs['content_type'] = 'application/json'
        kwargs['user_agent'] = f'imbi/{version} (GitLabClient)'
        self.logger.debug('%s %s', method, url)
        return await super().http_fetch(str(url), method=method, **kwargs)

    async def fetch_all_pages(self, *path, **query) -> typing.List[dict]:
        url = self.token.integration.api_endpoint
        for component in path:
            url /= str(component)
        if query:
            url = url.with_query(query)

        entries = []
        while url is not None:
            response = await self.api(url)
            if not response.ok:
                raise errors.InternalServerError(
                    'GET %s failed: %s', url, response.code,
                    title='GitLab API Failure')

            entries.extend(response.body)
            for link in response.links:
                if link['rel'] == 'next':
                    url = yarl.URL(link['target'])
                    break
            else:
                url = None
        return entries

    async def fetch_group(self, *group_path) -> typing.Optional[dict]:
        slug = urllib.parse.quote('/'.join(group_path), safe='')
        url = self.token.integration.api_endpoint
        path = f'{url.path.rstrip("/")}/groups/{slug}'
        url = url.with_path(path, encoded=True)
        response = await self.api(url)
        if response.code == 404:
            self.logger.debug('Group not found: %s', url)
            return None
        if response.ok:
            return response.body
        raise errors.InternalServerError(
            'failed to fetch group %s: %s', slug, response.code,
            title='GitLab API Failure')

    async def fetch_project(self, project_id: int) \
            -> typing.Optional[ProjectInfo]:
        url = (self.token.integration.api_endpoint / 'projects' /
               str(project_id))
        response = await self.api(url)
        if response.code == 404:
            return None
        if response.ok:
            return ProjectInfo.parse_obj(response.body)
        raise errors.InternalServerError(
            'failed to fetch project %s: %s', project_id, response.code,
            title='GitLab API Failure')

    async def create_project(self,
                             parent,
                             project_name: str,
                             **attributes) -> ProjectInfo:
        if self.restrict_to_user:
            parent = await self.fetch_user_namespace()

        for name, value in PROJECT_DEFAULTS.items():
            attributes.setdefault(name, value)
        attributes.update({
            'name': project_name,
            'namespace_id': parent['id'],
        })
        response = await self.api('projects', method='POST', body=attributes)
        if response.ok:
            return ProjectInfo.parse_obj(response.body)
        raise errors.InternalServerError(
            'failed to create project %s: %s', project_name, response.code,
            title='GitLab API Failure', gitlab_response=response.body)

    async def commit_tree(self,
                          project_info: ProjectInfo,
                          project_dir: pathlib.Path,
                          commit_message: str):
        self.logger.info('creating commit for %s', project_info.id)

        files = [path for path in project_dir.glob('**/*') if path.is_file()]
        self.logger.debug('found %d files in %s', len(files), project_dir)

        actions = [
            {
                'action': 'create',
                'content': base64.b64encode(file.read_bytes()).decode('ascii'),
                'encoding': 'base64',
                'file_path': str(file.relative_to(project_dir)),
            }
            for file in files
        ]
        actions.extend([
            {
                'action': 'chmod',
                'execute_filemode': True,
                'file_path': str(file.relative_to(project_dir)),
            }
            for file in files
            if file.stat().st_mode & stat.S_IXUSR
        ])
        self.logger.debug('creating commit with %d actions', len(actions))

        project_url = yarl.URL(project_info.links.self)
        response = await self.api(
            project_url / 'repository' / 'commits',
            method='POST',
            body={
                'branch': project_info.default_branch,
                'commit_message': commit_message,
                'actions': actions,
            },
        )
        if response.ok:
            return response.body
        raise errors.InternalServerError(
            'failed to commit to %s: %s', project_url, response.code,
            title='GitLab API Failure', gitlab_response=response.body)

    async def fetch_user_information(self):
        if self._user_info:
            return self._user_info
        response = await self.api(self.token.integration.api_endpoint / 'user')
        if response.ok:
            self._user_info = response.body
            return self._user_info
        raise errors.InternalServerError(
            'failed to retrieve GitLab user information: %s', response.code,
            title='GitLab API Failure')

    async def fetch_user_namespace(self):
        if self._user_namespace:
            return self._user_namespace
        user_info = await self.fetch_user_information()
        response = await self.api(
            self.token.integration.api_endpoint /
            'namespaces' / user_info['username'])
        if response.ok:
            self._user_namespace = response.body
            return self._user_namespace
        raise errors.InternalServerError(
            'failed to retrieve GitLab user namespace for %s: %s',
            user_info['username'], response.code, title='GitLab API Failure')



"""
Request Handler for the Change Log

"""
from tornado import web

from imbi.endpoints import base


class RequestHandler(base.RequestHandler):

    NAME = 'operations-change-log'

    @web.authenticated
    def get(self, *args, **kwargs):
        return self.send_response([])



"""
Async OpenSearch Client Wrapper
===============================

Wraps OpenSearch client operations to use Redis as queue to perform indexing
operations asynchronously.

"""
import asyncio
import datetime
import decimal
import json
import logging
import re
import typing
import uuid

import aioredis
import opensearchpy

LOGGER = logging.getLogger(__name__)


def normalize(value_in: dict) -> dict:
    for key, value in value_in.items():
        if isinstance(value, dict):
            value_in[key] = normalize(value)
        elif isinstance(value, datetime.date):
            value_in[key] = value.isoformat()
        elif isinstance(value, datetime.datetime):
            value_in[key] = value.replace(microsecond=0).isoformat()
        elif isinstance(value, decimal.Decimal):
            value_in[key] = str(value_in[key])
        elif isinstance(value, uuid.UUID):
            value_in[key] = str(value)
        elif value == '':
            value_in[key] = None
    return value_in


def sanitize_key(value: str) -> str:
    key = value.lower().replace(' ', '_').replace('/', '_')
    return re.sub('_+', '_', key)


def sanitize_keys(value: dict) -> dict:
    for key in list(value.keys()):
        sanitized = sanitize_key(key)
        if key != sanitized:
            value[sanitized] = value[key]
            del value[key]
    for key in value.keys():
        if isinstance(value[key], dict):
            value[key] = sanitize_keys(value[key])
    return value


class OpenSearch:

    PROCESS_DELAY = 5
    PENDING_KEY = 'documents-pending'

    def __init__(self, settings: dict):
        self.client: typing.Optional[opensearchpy.AsyncOpenSearch] = None
        self.loop: typing.Optional[asyncio.AbstractEventLoop] = None
        self.redis: typing.Optional[aioredis.Redis] = None
        self.settings = settings
        self.timer_handle: typing.Optional[asyncio.TimerHandle] = None

    async def initialize(self) -> bool:
        self.loop = asyncio.get_running_loop()

        try:
            self.redis = aioredis.Redis(await aioredis.create_pool(
                self.settings.get('redis_url', 'redis://localhost:6379/2')))
        except (OSError, ConnectionRefusedError) as error:
            LOGGER.info('Error connecting to OpenSearch redis: %r', error)
            return False

        self.client = opensearchpy.AsyncOpenSearch(
            **self.settings['connection'])

        self.timer_handle = self.loop.call_soon(
            lambda: asyncio.ensure_future(self._process()))
        return True

    async def create_index(self, index: str) -> bool:
        LOGGER.debug('Creating %s index in OpenSearch', index)
        try:
            await self.client.indices.create(index)
        except opensearchpy.exceptions.RequestError as err:
            if 'resource_already_exists_exception' not in err.error:
                LOGGER.warning('Index creation error: %r', err)
                return False
        return True

    async def create_mapping(self, index: str, mappings: dict) -> bool:
        try:
            await self.client.indices.put_mapping(
                index=index, body={'properties': mappings})
        except opensearchpy.exceptions.RequestError as err:
            LOGGER.debug('Mapping update failure: %r', err)
            return False
        return True

    async def delete_document(self, index: str, document_id: str) -> None:
        LOGGER.debug('Deleting %s from %s', document_id, index)
        try:
            await self.client.delete(index, document_id)
        except (
            opensearchpy.exceptions.NotFoundError,
            opensearchpy.exceptions.RequestError
        ) as err:
            LOGGER.warning('Deletion of %s:%s failed: %r',
                           index, document_id, err)

    async def documents_pending(self) -> int:
        """Return the number of documents pending indexing"""
        return await self.redis.scard(self.PENDING_KEY)

    async def index_document(self,
                             index: str,
                             document_id: str,
                             document: dict,
                             sync: bool = False) -> None:
        """Queue a document to be added the OpenSearch index"""
        LOGGER.debug('Queueing %s:%s to be indexed', index, document_id)
        if not sync:
            await self.redis.set(
                f'{index}:{document_id}',
                json.dumps(normalize(sanitize_keys(document)), indent=0))
            await self.redis.sadd(self.PENDING_KEY, f'{index}:{document_id}')
            return
        await self._index_document(
            index, document_id, normalize(sanitize_keys(document)))

    async def search(self, index: str, query: str, max_results: int = 1000) \
            -> typing.Dict[str, typing.List[dict]]:
        result = await self.client.search(
            body={
                'query': {
                    'query_string': {'query': query, 'size': max_results}}},
            index=index)
        return {'hits': [r['_source'] for r in result['hits']['hits']]}

    async def stop(self) -> None:
        if self.timer_handle and not self.timer_handle.cancelled():
            self.timer_handle.cancel()
        if self.client is not None:
            await self.client.close()

    async def _process(self) -> None:
        if await self._process_document():
            self.timer_handle = self.loop.call_soon(
                lambda: asyncio.ensure_future(self._process()))
            return
        self.timer_handle = self.loop.call_later(
            self.PROCESS_DELAY, lambda: asyncio.ensure_future(self._process()))

    async def _process_document(self) -> bool:
        """Index a single document in the index, if any are queued"""
        key = await self.redis.spop(self.PENDING_KEY)
        if not key:
            return False

        index, document_id = key.decode('utf-8').split(':')
        LOGGER.debug('Processing %s:%s', index, document_id)
        document = await self.redis.get(key)
        if not document:
            LOGGER.warning('Failed to load %s from redis', key)
            return False

        if not await self._index_document(
                index, document_id, json.loads(document)):
            return False

        result = await asyncio.gather(
            self.redis.delete(key),
            self.redis.scard(self.PENDING_KEY))
        LOGGER.debug(
            'Processing of %s:%s is complete with %i documents pending',
            index, document_id, result[1])
        return True

    async def _index_document(self,
                              index: str,
                              document_id: str,
                              document: dict) -> bool:
        """Invoked to index a document in ElasticSearch"""
        try:
            await self.client.index(
                index, body=document, id=document_id)
        except opensearchpy.RequestError as err:
            LOGGER.warning('Failed to index %s in %s: %s',
                           document_id, index, err)
            return False
        return True



from . import base


class RequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'permissions'

    @base.require_permission('admin')
    async def get(self):
        self.send_response(list(self.settings['permissions']))



"""
Imbi
====

Imbi is an operational management platform for medium to large environments.

"""
import warnings

import pkg_resources

version = pkg_resources.get_distribution('imbi').version
warnings.simplefilter('ignore', UserWarning)



from __future__ import annotations

import datetime
import decimal
import typing

import dateutil.parser


def coerce_project_fact_values(
        rows: list[dict[str, typing.Any]]
) -> list[dict[str, None | bool | decimal.Decimal | int | str]]:
    """Shared method to coerce the values to the correct data types in the
    output for project facts

    """
    for row in rows:
        row['value'] = coerce_project_fact(row['data_type'], row['value'])
        del row['data_type']

    return rows


def coerce_project_fact(
        data_type: str, value: typing.Any
) -> None | bool | decimal.Decimal | int | str:
    """Coerce a single fact based on its data type."""
    if not value and value != 0:
        value = None
    elif data_type == 'boolean':
        value = 'false' if value is None else str(value).lower()
        if value in ('y', 'yes', 't', 'true', 'on', '1'):
            value = True
        elif value in ('n', 'no', 'f', 'false', 'off', '0'):
            value = False
        else:
            raise ValueError(f'{value!r} is not a valid Boolean')
    elif data_type == 'decimal':
        try:
            value = decimal.Decimal(value)
        except decimal.InvalidOperation as error:
            raise ValueError(f'{value!r} is not a valid decimal: {error}')
    elif data_type == 'integer':
        try:
            value = int(value)
        except TypeError as error:
            raise ValueError(f'{value!r} is not a valid integer: {error}')
    elif data_type == 'date':
        value = dateutil.parser.parse(str(value))
        value = value.date().isoformat()
    elif data_type == 'timestamp':
        value = dateutil.parser.parse(str(value))
        if value.tzinfo is None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        value = value.isoformat()
    elif data_type == 'string':
        value = str(value)
    else:
        raise ValueError(f'{data_type!r} is not a known fact type')

    return value



"""
Base Request Handlers

"""
import asyncio
import datetime
import logging
import typing
import uuid
from email import utils

import jsonpatch
import problemdetails
import sprockets_postgres as postgres
from openapi_core.deserializing.exceptions import DeserializeError
from openapi_core.schema.media_types.exceptions import InvalidContentType
from openapi_core.templating.paths.exceptions import \
    OperationNotFound, PathNotFound
from openapi_core.unmarshalling.schemas.exceptions import ValidateError
from openapi_core.validation.exceptions import InvalidSecurity
from sprockets.http import mixins
from sprockets.mixins import mediatype
from tornado import httputil, web

from imbi import cors, errors, session, user, version

LOGGER = logging.getLogger(__name__)


def require_permission(permission):
    """Decorator function for requiring a permission string for an endpoint

    :param str permission: The permission string to require
    :raises: problemdetails.Problem

    """
    def _require_permission(f):
        def wrapped(self, *args, **kwargs):
            """Inner-wrapping of the decorator that performs the logic"""
            if not self._current_user or \
                    not self._current_user.has_permission(permission):
                if self._respond_with_html:
                    return self.render(
                        'index.html',
                        javascript_url=self.application.settings.get(
                            'javascript_url'))
                raise errors.Forbidden('%r does not have the "%s" permission',
                                       self._current_user, permission)
            return f(self, *args, **kwargs)
        return wrapped
    return _require_permission


class RequestHandler(cors.CORSMixin,
                     postgres.RequestHandlerMixin,
                     mixins.ErrorLogger,
                     problemdetails.ErrorWriter,
                     mediatype.ContentMixin,
                     web.RequestHandler):
    """Base RequestHandler class used for recipients and subscribers."""

    APPLICATION_JSON = 'application/json'
    TEXT_HTML = 'text/html'
    NAME = 'Base'
    ITEM_NAME = ''

    def __init__(self,
                 application,
                 request: httputil.HTTPServerRequest,
                 **kwargs):
        super().__init__(application, request, **kwargs)
        self.logger = logging.getLogger(f'imbi.endpoints.{self.NAME}')
        self.session: typing.Optional[session.Session] = None
        self._current_user: typing.Optional[user.User] = None
        self._links = {}

    async def prepare(self) -> None:
        """Prepare the request handler for the request. If the application
        is not ready return a ``503`` error.

        Checks for a session cookie and if present, loads the session into
        the current user and authenticates it. If authentication fails,
        the current user and cookie is cleared.

        """
        if not self.application.ready_to_serve:
            return self.send_error(503, reason='Application not ready')
        self.session = session.Session(self)
        await self.session.initialize()
        self._current_user = await self.get_current_user()
        future = super().prepare()
        if asyncio.isfuture(future) or asyncio.iscoroutine(future):
            await future

    def on_finish(self) -> None:
        """Invoked after a request has completed"""
        super().on_finish()
        self.application.loop.add_callback(
            self.application.stats.incr,
            {
                'key': 'http_requests',
                'endpoint': self.NAME,
                'method': self.request.method,
                'status': self.get_status()
            })

        self.application.loop.add_callback(
            self.application.stats.add_duration,
            {
                'key': 'http_request_duration',
                'endpoint': self.NAME,
                'method': self.request.method,
                'status': self.get_status()
            },
            self.request.request_time())

    def compute_etag(self) -> None:
        """Override Tornado's built-in ETag generation"""
        return None

    async def get_current_user(self) -> typing.Optional['user.User']:
        """Used by the system to manage authentication behaviors"""
        if self.session and self.session.user:
            return self.session.user
        token = self.request.headers.get('Private-Token', None)
        if token:
            current_user = user.User(self.application, token=token)
            if await current_user.authenticate():
                return current_user

    def get_template_namespace(self) -> dict:
        """Returns a dictionary to be used as the default template namespace.

        The results of this method will be combined with additional defaults
        in the :mod:`tornado.template` module and keyword arguments to
        :meth:`~tornado.web.RequestHandler.render`
        or :meth:`~tornado.web.RequestHandler.render_string`.

        """
        namespace = super(RequestHandler, self).get_template_namespace()
        namespace.update({'version': version})
        return namespace

    def on_postgres_timing(self,
                           metric_name: str,
                           duration: float) -> None:
        """Invoked by sprockets-postgres after each query"""
        self.application.loop.add_callback(
            self.application.stats.add_duration,
            {
                'key': 'postgres_query_duration',
                'query': metric_name,
                'endpoint': self.NAME
            },
            duration)

    def send_response(self, value: typing.Union[dict, list]) -> None:
        """Send the response to the client"""
        if 'self' not in self._links:
            self._add_self_link(self.request.path)
            self._add_link_header()
        if hasattr(self, 'TTL') and \
                not self.request.headers.get('Pragma') == 'no-cache':
            self._add_response_caching_headers(self.TTL)
        super().send_response(value)

    def set_default_headers(self) -> None:
        """Override the default headers, setting the Server response header"""
        super().set_default_headers()
        self.set_header('Server', self.settings['server_header'])

    def write_error(self, status_code, **kwargs):
        if self._respond_with_html:
            return self.render(
                'error.html',
                javascript_url=self.application.settings.get('javascript_url'),
                status_code=status_code, **kwargs)
        super().write_error(status_code, **kwargs)

    def _add_last_modified_header(self, value: datetime.datetime) -> None:
        """Add a RFC-822 formatted timestamp for the Last-Modified HTTP
        response header.

        """
        if not value:
            return
        self.set_header('Last-Modified', self._rfc822_date(value))

    def _add_link_header(self) -> None:
        """Takes the accumulated links and creates a link header value"""
        links = []
        for rel, path in self._links.items():
            links.append('<{}://{}{}>; rel="{}"'.format(
                self.request.protocol, self.request.host, path, rel))
        if links:
            self.add_header('Link', ','.join(links))

    def _add_self_link(self, path: str) -> None:
        """Adds the self Link response header"""
        self._links['self'] = path

    def _add_response_caching_headers(self, ttl: int) -> None:
        """Adds the cache response headers for the object being returned."""
        self.add_header('Cache-Control', 'public, max-age={}'.format(ttl))

    @property
    def _respond_with_html(self) -> bool:
        """Returns True if the current response should respond with HTML"""
        return self.get_response_content_type().startswith(self.TEXT_HTML)

    @staticmethod
    def _rfc822_date(value: datetime.datetime) -> str:
        """Return an RFC-822 formatted timestamp for the given value"""
        return utils.format_datetime(value)


class AuthenticatedRequestHandler(RequestHandler):
    """RequestHandler base class for authenticated requests"""

    async def prepare(self) -> None:
        await super().prepare()

        # Don't require authorization for pre-flight requests
        if self.cors.is_preflight:
            return

        if not self._current_user:
            if self._respond_with_html:
                return await self.render(
                    'index.html',
                    javascript_url=self.application.settings.get(
                        'javascript_url'))
            self.set_status(401)
            await self.finish()
            raise web.Finish()


class ValidatingRequestHandler(AuthenticatedRequestHandler):
    """Validates the request against the OpenAPI spec"""
    async def prepare(self) -> None:
        await super().prepare()

        # Don't validate preflight OPTIONS requests... otherwise,
        # we need to include the OPTIONS stanza in every OpenAPI
        # path spec
        if self.cors.is_preflight:
            return

        try:
            self.application.validate_request(self.request)
        except DeserializeError as err:
            raise errors.BadRequest('Failed to deserialize body: %s', err,
                                    detail=str(err))
        except InvalidSecurity as err:
            raise errors.InternalServerError(
                'Invalid OpenAPI spec security: %s', err,
                title='OpenAPI Security Error')
        except OperationNotFound as err:
            raise errors.MethodNotAllowed(err.method.upper())
        except InvalidContentType as err:
            raise errors.UnsupportedMediaType(err.mimetype)
        except PathNotFound as err:
            raise errors.InternalServerError('OpenAPI Spec Error: %s', err,
                                             title='OpenAPI Spec Error',
                                             detail=str(err))
        except ValidateError as err:
            raise errors.BadRequest('Request failed to validate: %s', err,
                                    detail='The request did not validate',
                                    errors=[str(e).split('\n')[0]
                                            for e in err.schema_errors])


class CRUDRequestHandler(ValidatingRequestHandler):
    """CRUD request handler to reduce large amounts of duplicated code"""

    NAME = 'default'
    DEFAULTS = {}
    ID_KEY: typing.Union[str, list] = 'id'
    IS_COLLECTION = False
    FIELDS = None
    ITEM_NAME = None  # Used to create link headers for POST requests
    TTL = 300

    DELETE_SQL: typing.Optional[str] = None
    GET_SQL: typing.Optional[str] = None
    PATCH_SQL: typing.Optional[str] = None
    POST_SQL: typing.Optional[str] = None

    async def delete(self, *args, **kwargs):
        if self.DELETE_SQL is None:
            self.logger.debug('DELETE_SQL not defined')
            raise errors.MethodNotAllowed('DELETE')
        await self._delete(kwargs)

    async def get(self, *args, **kwargs):
        if self.GET_SQL is None:
            self.logger.debug('GET_SQL not defined')
            raise errors.MethodNotAllowed('GET')
        if self._respond_with_html:
            return self.render(
                'index.html',
                javascript_url=self.application.settings.get('javascript_url'))
        await self._get(kwargs)

    async def patch(self, *args, **kwargs):
        if self.PATCH_SQL is None:
            self.logger.debug('PATCH_SQL not defined')
            raise errors.MethodNotAllowed('PATCH')
        await self._patch(kwargs)

    async def post(self, *args, **kwargs):
        if self.POST_SQL is None:
            self.logger.debug('POST_SQL not defined')
            raise errors.MethodNotAllowed('POST')
        await self._post(kwargs)

    def send_response(self, value: typing.Union[dict, list]) -> None:
        """Send the response to the client"""
        if isinstance(value, list):
            return super().send_response(value)

        if not (self.request.method == 'GET' and self.IS_COLLECTION):
            self._add_last_modified_header(
                value.get('last_modified_at', value.get('created_at')))
            for key in {'created_at', 'last_modified_at'}:
                if key in value:
                    del value[key]
            if self.ID_KEY:
                if isinstance(self.ID_KEY, list):
                    args = [str(value[k]) for k in self.ID_KEY]
                else:
                    args = [str(value[self.ID_KEY])]

                try:
                    self._add_self_link(
                        self.reverse_url(self.ITEM_NAME or self.NAME, *args))
                except (AssertionError, KeyError):
                    self.logger.debug('Failed to reverse URL for %s %r',
                                      self.NAME, args)
            self._add_link_header()
        super().send_response(value)

    async def _delete(self, kwargs):
        result = await self.postgres_execute(
            self.DELETE_SQL, self._get_query_kwargs(kwargs),
            'delete-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)
        self.set_status(204, reason='Item Deleted')

    async def _get(self, kwargs):
        result = await self.postgres_execute(
            self.GET_SQL, self._get_query_kwargs(kwargs),
            'get-{}'.format(self.NAME))
        if not result.row_count or not result.row:
            raise errors.ItemNotFound(instance=self.request.uri)
        for key, value in result.row.items():
            if isinstance(value, uuid.UUID):
                result.row[key] = str(value)
        self.send_response(result.row)

    def _get_query_kwargs(self, kwargs) -> dict:
        if isinstance(self.ID_KEY, list):
            return {k: kwargs[k] for k in self.ID_KEY}
        return {self.ID_KEY: kwargs[self.ID_KEY]}

    async def _patch(self, kwargs):
        patch_value = self.get_request_body()

        result = await self.postgres_execute(
            self.GET_SQL, self._get_query_kwargs(kwargs),
            'get-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)

        original = dict(result.row)
        for key in {'created_at', 'created_by',
                    'last_modified_at', 'last_modified_by'}:
            if key in original:
                del original[key]

        for key, value in original.items():
            if isinstance(value, uuid.UUID):
                original[key] = str(value)

        # Apply the patch to the current value
        patch = jsonpatch.JsonPatch(patch_value)
        updated = patch.apply(original)

        # Bail early if there are no changes
        if not {k: original[k] for k in original
                if k in updated and original[k] != updated[k]}:
            self._add_self_link(self.request.path)
            self._add_link_header()
            return self.set_status(304)

        if isinstance(self.ID_KEY, list):
            for key in self.ID_KEY:
                updated['current_{}'.format(key)] = kwargs[key]
        else:
            updated['current_{}'.format(self.ID_KEY)] = kwargs[self.ID_KEY]
        updated['username'] = self._current_user.username

        result = await self.postgres_execute(
            self.PATCH_SQL, updated,
            'patch-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.DatabaseError('No rows returned from PATCH_SQL',
                                       title='Failed to update record')

        # Send the new record as a response
        await self._get(self._get_query_kwargs(updated))

    async def _post(self, kwargs) -> dict:
        values = self.get_request_body()

        # Handle compound keys for child object CRUD
        if isinstance(self.ID_KEY, list):
            for key in self.ID_KEY:
                if key not in values and key in kwargs:
                    values[key] = kwargs[key]
        elif self.ID_KEY not in values and self.ID_KEY in kwargs:
            values[self.ID_KEY] = kwargs[self.ID_KEY]

        # Set defaults of None for all fields in insert
        for name in self.FIELDS:
            if name not in values:
                values[name] = self.DEFAULTS.get(name)

        values['username'] = self._current_user.username
        result = await self.postgres_execute(
            self.POST_SQL, values, 'post-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.DatabaseError('No rows returned from POST_SQL',
                                       title='Failed to create record')

            # Return the record as if it were a GET
        await self._get(self._get_query_kwargs(result.row))
        return result.row


class CollectionRequestHandler(CRUDRequestHandler):

    DEFAULTS = {}
    ID_KEY: typing.Union[str, list] = 'id'
    IS_COLLECTION: True
    FIELDS = None
    ITEM_NAME = None  # Used to create link headers for POST requests
    COLLECTION_SQL = """SELECT * FROM pg_tables WHERE schemaname = 'v1';"""
    TTL = 300

    async def get(self, *args, **kwargs):
        result = await self.postgres_execute(
            self.COLLECTION_SQL, kwargs,
            metric_name='get-{}'.format(self.NAME))
        self.send_response(result.rows)


class AdminCRUDRequestHandler(CRUDRequestHandler):

    @require_permission('admin')
    async def delete(self, *args, **kwargs):
        await super().delete(*args, **kwargs)

    @require_permission('admin')
    async def get(self, *args, **kwargs):
        await super().get(*args, **kwargs)

    @require_permission('admin')
    async def patch(self, *args, **kwargs):
        await super().patch(*args, **kwargs)

    @require_permission('admin')
    async def post(self, *args, **kwargs):
        await super().post(*args, **kwargs)



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = ['id']
    FIELDS = ['id', 'fact_type_id', 'min_value', 'max_value',  'score']

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, fact_type_id, created_at, created_by,
               last_modified_at, last_modified_by, min_value, max_value, score
          FROM v1.project_fact_type_ranges
         WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'fact-type-ranges'
    ITEM_NAME = 'fact-type-range'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT id, fact_type_id, min_value, max_value, score
          FROM v1.project_fact_type_ranges
      ORDER BY id""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.project_fact_type_ranges
                    (fact_type_id, created_by, min_value, max_value, score)
             VALUES (%(fact_type_id)s,  %(username)s, %(min_value)s,
                     %(max_value)s, %(score)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'fact-type-range'

    DELETE_SQL = 'DELETE FROM v1.project_fact_type_ranges WHERE id=%(id)s'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.project_fact_type_ranges
           SET fact_type_id=%(fact_type_id)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               min_value=%(min_value)s,
               max_value=%(max_value)s,
               score=%(score)s
         WHERE id=%(id)s""")



import re

from tornado import web

from imbi.endpoints import base


class RequestHandler(base.ValidatingRequestHandler):

    NAME = 'activity-feed'

    SQL = re.sub(r'\s+', ' ', """\
        SELECT "when", namespace_id, namespace, project_id, project_name,
                project_type, who, display_name, email_address, what
          FROM v1.activity_feed
         ORDER BY "when" DESC
        OFFSET {offset}
         LIMIT {limit};""")

    @web.authenticated
    async def get(self):
        result = await self.postgres_execute(
            self.SQL.format(
                limit=int(self.get_query_argument('limit', '25')),
                offset=int(self.get_query_argument('offset', '0'))),
            metric_name='reports-compliance')
        self.send_response(result.rows)



from imbi.endpoints import base


class UserRequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'ui-user'

    def get(self, *args, **kwargs):
        user = self.current_user.as_dict()
        del user['password']
        self.send_response(user)



from __future__ import annotations

import logging

import pydantic
import sprockets.mixins.http
import yarl

from imbi import errors, version


class _ProjectCreationResponse(pydantic.BaseModel):
    name: str
    slug: str
    id: str


class _ProjectKeysDsn(pydantic.BaseModel):
    public: pydantic.HttpUrl
    secret: pydantic.HttpUrl


class _ProjectKeysResponse(pydantic.BaseModel):
    dsn: _ProjectKeysDsn
    id: str
    isActive: bool
    projectId: int


class ProjectInfo(pydantic.BaseModel):
    name: str
    slug: str
    id: str
    link: pydantic.HttpUrl
    keys: dict[str, str]


class SentryClient(sprockets.mixins.http.HTTPClientMixin):
    def __init__(self, application) -> None:
        super().__init__()
        self.logger = logging.getLogger(__package__).getChild('SentryClient')

        settings = application.settings['automations']['sentry']
        self.url = yarl.URL(settings['url'])
        self.api_url = self.url / 'api' / '0'
        self.enabled = settings['enabled']
        self.organization = settings.get('organization')
        self.token = settings.get('auth_token')
        self.headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.token}',
        }

    async def api(self, url: yarl.URL | str, *,
                  method: str = 'GET', **kwargs):
        if not self.enabled:
            raise errors.InternalServerError('Sentry client is not enabled',
                                             title='Sentry Client Error')

        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            url = self.api_url / url.path.lstrip('/')

        # the sentry API wants a trailing slash
        url = str(url).rstrip('/') + '/'

        request_headers = kwargs.pop('request_headers', {})
        request_headers.update(self.headers)
        if kwargs.get('body') is not None:
            request_headers['Content-Type'] = 'application/json'
            kwargs['content_type'] = 'application/json'
        kwargs['request_headers'] = request_headers
        kwargs['user_agent'] = f'imbi/{version} (SentryClient)'
        rsp = await super().http_fetch(str(url), method=method, **kwargs)
        if not rsp.ok:
            self.logger.error('%s %s failed: %s', method, url, rsp.code)
        return rsp

    async def create_project(self, team_slug: str,
                             project_name: str) -> ProjectInfo:
        url = yarl.URL('/teams') / self.organization / team_slug / 'projects'
        response = await self.api(url, method='POST',
                                  body={'name': project_name})
        if not response.ok:
            self.logger.error('Sentry API failure: body %r', response.body)
            raise errors.InternalServerError(
                'POST %s failed: %s', url, response.code,
                title='Sentry API Failure')
        else:
            try:
                data = _ProjectCreationResponse.parse_obj(response.body)
            except pydantic.ValidationError as error:
                self.logger.error('failed to parse response body: %r',
                                  response.body)
                raise errors.InternalServerError(
                    'Failed to parse sentry create_project response: %s',
                    error, title='Sentry Client Failure')

        project = ProjectInfo(
            id=data.id,
            name=data.name,
            slug=data.slug,
            link=str(self.url / 'organizations' / self.organization /
                     'projects' / data.slug),
            keys={}
        )

        # TODO update the project platform based on the project details
        #      so that python projects use "platform=python" and
        #      JS projects use "platform=javascript"
        # Not sure if this is actually necessary so I'm omitting it.
        # https://github.com/AWeber-Imbi/imbi/issues/73
        # ref: https://docs.sentry.io/api/projects/update-a-project/
        # ref: https://docs.sentry.io/platforms/

        url = yarl.URL('/projects') / self.organization / project.slug

        try:
            response = await self.api(url / 'keys')
            if not response.ok:
                raise errors.InternalServerError(
                    'Failed to retrieve sentry project keys: %s',
                    response.code, title='Sentry API Error')
            try:
                data = _ProjectKeysResponse.parse_obj(response.body[0])
            except pydantic.ValidationError as error:
                self.logger.error('failed to parse response body: %r',
                                  response.body)
                raise errors.InternalServerError(
                    'Failed to parse sentry project keys response: %s',
                    error, title='Sentry Client Failure')
            else:
                project.keys.update({
                    'public': data.dsn.public,
                    'secret': data.dsn.secret,
                })

        # If we fail to fetch or extract the keys that the client
        # needs, then remove the project since we cannot use it.
        except errors.ApplicationError as error:
            self.logger.error('removing Sentry project %s due to error: %s',
                              project.slug, error)
            await self.remove_project(project.slug)
            raise

        return project

    async def remove_project(self, project_slug: str) -> None:
        await self.api(
            yarl.URL('/projects') / self.organization / project_slug,
            method='DELETE')



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'name'
    FIELDS = ['name', 'description', 'icon_class']
    DEFAULTS = {'icon_class': 'fas fa-mountain'}

    GET_SQL = re.sub(r'\s+', ' ', """\
        SELECT "name", created_at, created_by,
               last_modified_at, last_modified_by,
               description, icon_class
          FROM v1.environments
         WHERE "name"=%(name)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'environments'
    ITEM_NAME = 'environment'

    COLLECTION_SQL = re.sub(r'\s+', ' ', """\
        SELECT "name", description, icon_class
          FROM v1.environments
         ORDER BY "name" ASC""")

    POST_SQL = re.sub(r'\s+', ' ', """\
        INSERT INTO v1.environments
                    ("name", created_by, description, icon_class)
             VALUES (%(name)s, %(username)s, %(description)s, %(icon_class)s)
          RETURNING "name";""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'environment'

    DELETE_SQL = 'DELETE FROM v1.environments WHERE "name"=%(name)s'

    PATCH_SQL = re.sub(r'\s+', ' ', """\
        UPDATE v1.environments
           SET "name"=%(name)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               description=%(description)s,
               icon_class=%(icon_class)s
         WHERE "name"=%(current_name)s""")



import asyncio
import dataclasses
import datetime
import decimal
import logging
import re
import typing

from imbi import common, errors
if typing.TYPE_CHECKING:
    from imbi import app

decimal.getcontext().prec = 2
LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class CookieCutter:
    name: str
    project_type_id: int
    url: str


@dataclasses.dataclass
class Namespace:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: str
    last_modified_by: typing.Optional[datetime.datetime]
    name: str
    slug: str
    icon_class: str
    maintained_by: typing.Optional[typing.List[str]]
    gitlab_group_name: str
    sentry_team_slug: typing.Optional[str]

    SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               name,
               slug,
               icon_class,
               maintained_by,
               gitlab_group_name,
               sentry_team_slug
          FROM v1.namespaces
         WHERE id=%(id)s""")


@dataclasses.dataclass
class ProjectFact:
    id: int
    name: str
    recorded_at: datetime.datetime
    recorded_by: str
    value: typing.Union[
        bool,
        datetime.date,
        datetime.datetime,
        decimal.Decimal,
        int,
        None,
        str
    ]
    fact_type: str
    data_type: str
    description: typing.Optional[str]
    ui_options: typing.Optional[typing.List[str]]
    score: decimal.Decimal
    weight: int

    COLLECTION_SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        WITH project_type_id AS (SELECT project_type_id AS id
                                   FROM v1.projects
                                  WHERE id = %(obj_id)s)
        SELECT a.id,
               a.name,
               b.recorded_at,
               b.recorded_by,
               b.value,
               a.fact_type,
               a.data_type,
               a.description,
               a.ui_options,
               CASE WHEN b.value IS NULL THEN 0
                    ELSE CASE WHEN a.fact_type = 'enum' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_enums
                                           WHERE fact_type_id = b.fact_type_id
                                             AND value = b.value)
                              WHEN a.fact_type = 'range' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_ranges
                                           WHERE fact_type_id = b.fact_type_id
                                             AND b.value::NUMERIC(9,2)
                                         BETWEEN min_value AND max_value)
                              ELSE 0
                          END
                END AS score,
               a.weight
          FROM v1.project_fact_types AS a
     LEFT JOIN v1.project_facts AS b
            ON b.fact_type_id = a.id
           AND b.project_id = %(obj_id)s
         WHERE (SELECT id FROM project_type_id) = ANY(a.project_type_ids)
      ORDER BY a.name""")

    def __post_init__(self):
        try:
            value = common.coerce_project_fact(self.data_type, self.value)
        except ValueError:
            pass
        else:
            self.__setattr__('value', value)


@dataclasses.dataclass
class ProjectType:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    name: str
    slug: str
    plural_name: str
    description: typing.Optional[str]
    icon_class: typing.Optional[str]
    environment_urls: bool
    gitlab_project_prefix: typing.Optional[str]

    SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               name,
               slug,
               plural_name,
               description,
               icon_class,
               environment_urls,
               gitlab_project_prefix
          FROM v1.project_types
         WHERE id=%(id)s""")


@dataclasses.dataclass
class ProjectLink:
    link_type_id: int
    link_type: str
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    icon_class: typing.Optional[str]
    url: str

    COLLECTION_SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        SELECT a.link_type_id,
               b.link_type,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.icon_class,
               a.url
          FROM v1.project_links AS a
          JOIN v1.project_link_types AS b
            ON b.id = a.link_type_id
         WHERE a.project_id = %(obj_id)s
         ORDER BY b.link_type""")


@dataclasses.dataclass
class ProjectURL:
    environment: str
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    icon_class: typing.Optional[str]
    url: str

    COLLECTION_SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        SELECT a.environment,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.icon_class,
               a.url
          FROM v1.project_urls AS a
          JOIN v1.environments AS b
            ON b.name = a.environment
         WHERE a.project_id = %(obj_id)s
         ORDER BY b.name""")


@dataclasses.dataclass
class Project:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[datetime.datetime]
    last_modified_by: typing.Optional[str]
    namespace: Namespace
    project_type: ProjectType
    name: str
    slug: str
    description: typing.Optional[str]
    environments: typing.Optional[typing.List[str]]
    archived: bool
    gitlab_project_id: typing.Optional[int]
    sentry_project_slug: typing.Optional[str]
    sonarqube_project_key: typing.Optional[str]
    pagerduty_service_id: typing.Optional[str]
    facts: typing.Dict[str, str]
    links: typing.Dict[str, str]
    urls: typing.Dict[str, str]
    project_score: int

    SQL: typing.ClassVar = re.sub(r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               namespace_id,
               project_type_id,
               name,
               slug,
               description,
               environments,
               archived,
               gitlab_project_id,
               sentry_project_slug,
               sonarqube_project_key,
               pagerduty_service_id,
               v1.project_score(id) AS project_score
          FROM v1.projects
         WHERE id=%(id)s""")


async def _load(model: dataclasses.dataclass, obj_id: int,
                application: 'app.Application') -> dataclasses.dataclass:

    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for project %s: %s',
                     obj_id, exc)
        raise errors.DatabaseError(
            f'Error loading {model.__class__.__name__}', error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(
            model.SQL, {'id': obj_id}, 'model-load')
        if result.row_count:
            return model(**result.row)


async def _load_collection(model: dataclasses.dataclass,
                           obj_id: int,
                           application: 'app.Application') \
        -> typing.List[dataclasses.dataclass]:

    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for collection %s: %s',
                     obj_id, exc)
        raise errors.DatabaseError(
            f'Error loading {model.__class__.__name__} '
            f'collection for {obj_id}', error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(
            model.COLLECTION_SQL, {'obj_id': obj_id},
            'collection-load')
        return [model(**row) for row in result.rows]


async def namespace(namespace_id: int,
                    application: 'app.Application') -> Namespace:
    return await _load(Namespace, namespace_id, application)


async def project(project_id: int,
                  application: 'app.Application') -> Project:

    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for project %s: %s',
                     project_id, exc)
        raise errors.DatabaseError(
            f'Error loading Project {project_id}', error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(
            Project.SQL, {'id': project_id}, 'project-model-load')
        if result.row_count:
            values = dict(result.row)
            result = await asyncio.gather(
                namespace(values['namespace_id'], application),
                project_type(values['project_type_id'], application),
                project_facts(project_id, application),
                project_links(project_id, application),
                project_urls(project_id, application))
            del values['namespace_id']
            del values['project_type_id']
            values.update({
                'namespace': result[0],
                'project_type': result[1],
                'facts': {value.name: value.value for value in result[2]},
                'links': {value.link_type: value.url for value in result[3]},
                'urls': {value.environment: value.url for value in result[4]}})
            return Project(**values)


async def project_facts(project_id: int,
                        application: 'app.Application') \
        -> typing.List[ProjectFact]:
    return await _load_collection(ProjectFact, project_id, application)


async def project_links(project_id: int,
                        application: 'app.Application') \
        -> typing.List[ProjectLink]:
    return await _load_collection(ProjectLink, project_id, application)


async def project_type(project_type_id: int,
                       application: 'app.Application') -> ProjectType:
    return await _load(ProjectType, project_type_id, application)


async def project_urls(project_id: int,
                       application: 'app.Application') \
        -> typing.List[ProjectURL]:
    return await _load_collection(ProjectURL, project_id, application)
